from typing import TypeVar

NumberType = TypeVar("NumberType", bound=int)

__all__ = ["increase"]
# Not pure
def increase(a_number: NumberType, amount: NumberType = 1):
    a_number = a_number + amount
    return a_number



from dino_seedwork_be.storage.uow import SuperDBSessionUser

from .mixins import BusinessRuleValidationMixin

__all__ = ["DomainService"]


class DomainService(BusinessRuleValidationMixin, SuperDBSessionUser):
    """
    Domain services carry domain knowledge that doesnt naturally fit entities and value objects.
    """



from dataclasses import field

from returns.result import safe

from dino_seedwork_be.domain.value_objects import ID, UUID
from dino_seedwork_be.exceptions import IllegalArgumentException
from dino_seedwork_be.logic import DomainAssertionConcern

__all__ = ["IdentifiedDomainObject"]


class IdentifiedDomainObject(DomainAssertionConcern):
    _id: ID = field(hash=True)

    def __eq__(self, obj):
        return self._id == obj.id

    def __init__(self):
        super().__init__()

    def identity(self):
        return self._id

    def raw_id(self) -> str:
        return self._id.get_raw()

    @safe
    def setId(self, id: UUID | ID | str):
        match id:
            case UUID() as id:
                self._id = ID(id)
            case ID() as id:
                self._id = id
            case str() as id:
                self._id = ID(id)
            case _:
                raise IllegalArgumentException("id is not in correct type")
        return "OK"



from pydantic import BaseModel

__all__ = ["BusinessRule"]


class BusinessRule(BaseModel):
    """This is a base class for implementing domain rules"""

    class Config:
        arbitrary_types_allowed = True

    # This is an error message that broken rule reports back
    __message: str = "Business rule is broken"

    def get_message(self) -> str:
        return self.__message

    def is_broken(self) -> bool:
        return False

    def __str__(self):
        return f"{self.__class__.__name__} {super().__str__()}"



import logging

__all__ = ["LOG_FORMAT", "SIMPLE_LOGGER"]

LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)

SIMPLE_LOGGER = logging.getLogger(__name__)



from typing import Generic, Type, TypeVar

from dino_seedwork_be.serializer.AbstractSerializer import AbstractSerializer
from dino_seedwork_be.serializer.Serializable import JSONSerializable

ObjectSerializable = TypeVar("ObjectSerializable", bound=JSONSerializable)
UnderlyingObject = TypeVar("UnderlyingObject")

__all__ = ["ObjectSerializer"]


class ObjectSerializer(AbstractSerializer, Generic[ObjectSerializable]):
    def __init__(self):
        super().__init__()

    def serialize(self, anObj: ObjectSerializable, toJson=False) -> str | object:
        if toJson:
            return str(self.json_marshaller().encode(anObj, unpicklable=False))
        return anObj.get_value()

    def deserialize(self, aJson, cls: Type[ObjectSerializable]):
        data = self.json_marshaller().decode(aJson)
        return cls.restore(data)



from sqlalchemy.ext.asyncio.session import AsyncSession

from dino_seedwork_be.storage.uow import DBSessionUser

__all__ = ["AlchemyRepository"]


class AlchemyRepository(DBSessionUser[AsyncSession]):
    pass



from dino_seedwork_be.domain import ID, Entity

__all__ = ["get_identity", "get_raw_identity"]


def get_identity(aEntity: Entity) -> ID:
    return aEntity.identity()


def get_raw_identity(aEntity: Entity) -> str:
    return aEntity.raw_id()



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from returns.pipeline import flow
from returns.pointfree import map_
from sqlalchemy.engine import Result

from dino_seedwork_be.logic import AssertionConcern
from dino_seedwork_be.utils import get_class_name, set_protected_attr, unwrap

IdRawType = TypeVar("IdRawType")

__all__ = ["AbstractIdentity"]


class AbstractIdentity(ABC, AssertionConcern, Generic[IdRawType]):

    _id: IdRawType

    def id(self) -> IdRawType:
        return self._id

    def __eq__(self, __o: object) -> bool:
        match __o:
            case AbstractIdentity():
                return type(__o) == type(self) and self.id() == __o.id()
            case _:
                return False

    def __str__(self) -> str:
        return f"{get_class_name(self)} [id={str(id)}]"

    def __init__(self, id: IdRawType) -> None:
        super().__init__()
        unwrap(self.set_id(id))

    @abstractmethod
    def validate(self, an_id: IdRawType) -> Result:
        pass

    def set_id(self, an_id: IdRawType):
        return flow(
            self.assert_argument_not_null(an_id),
            map_(self.validate),
            map_(lambda _: set_protected_attr(self, "_id", an_id)),
        )
